import pytest
import dbhydra.dbhydra_core as dh
import pandas as pd
import keepvariable.keepvariable_core as kv

import dbhydra.tests.test_cases as test_cases



db_details_mongo = kv.load_variable_safe("tests/test_mongo_db.kpv", varname="db_details")


@pytest.fixture(scope="module")
def mongodb():
    return dh.MongoDb(db_details=db_details_mongo)


def test_mongo_server_connection_and_table_presence(mongodb):
    with mongodb.connect_to_db():
        tables = mongodb.get_all_tables()

    assert "testing_table" in tables


@pytest.mark.parametrize(
    "db,compare_df, table_name",
    [
        ("mongodb", pd.DataFrame({"name": ["John"]}).astype("object"), "testing_table"),
        ("mongodb", pd.DataFrame([]).astype("object"), "testing_table_insert"),
    ],
)
def test_mongo_select(db, compare_df, table_name, request):
    db = request.getfixturevalue(db)
    with db.connect_to_db():
        tables = db.generate_table_dict()
        testing_table = tables[table_name]
        response_df = testing_table.select_to_df().astype("object")

    if "_id" in response_df.columns:
        response_df = response_df.drop(["_id"], axis=1)

    assert response_df.equals(compare_df)


@pytest.mark.parametrize(
    "db,input_df,compare_df,seq,",
    [
        ("mongodb", test_cases.test_case1[0], test_cases.test_case1[1], test_cases.test_case1[2]),
    ],
)
def test_mongo_crud(db, input_df, compare_df, seq, request):
    db = request.getfixturevalue(db)
    with db.connect_to_db():
        tables = db.generate_table_dict()
        testing_table = tables['testing_table_insert']
        testing_table.insert_from_df(input_df)
        seq(testing_table, {'age': 5}, {'name' : 'Lisa'}, {"$set": {"age":100}}, { "age": 144})

        resp_df = testing_table.select_to_df().astype('object')
        testing_table.delete()

    resp_df = resp_df.drop(['_id'], axis=1)

    assert resp_df.equals(compare_df)



import pandas as pd
import math
import json


class Migrator:
    """It was somewhat tested only for MySQL and Postgres dialect"""
    
    def __init__(self, db=None):
        self.db = db
        self.migration_number = 1
        self.migration_list = []

    def process_migration_dict(self, migration_dict):
        matching_table_class = self.db.matching_table_class #E.g. MysqlTable
        
        assert len(migration_dict.keys()) == 1
        operation = list(migration_dict.keys())[0]
        options = migration_dict[operation]
        if operation == "create":
            table = matching_table_class(self.db, options["table_name"], options["columns"], options["types"])
            table.convert_types_from_mysql()
            table.create()
        elif operation == "drop":
            table = matching_table_class(self.db, options["table_name"])
            table.drop()
        elif operation == "add_column":
            table = matching_table_class(self.db, options["table_name"])
            table.initialize_columns()
            table.initialize_types()
            table.convert_types_from_mysql()
            table.add_column(options["column_name"], options["column_type"])
        elif operation == "modify_column":
            table = matching_table_class(self.db, options["table_name"])
            table.initialize_columns()
            table.initialize_types()
            table.convert_types_from_mysql()
            table.modify_column(options["column_name"], options["column_type"])
        elif operation == "drop_column":
            table = matching_table_class(self.db, options["table_name"])
            table.initialize_columns()
            table.initialize_types()
            table.drop_column(options["column_name"])

    def next_migration(self):
        self.migration_number += 1
        self.migration_list = []

    def migrate(self, migration_list):
        for i, migration_dict in enumerate(migration_list):
            self.process_migration_dict(migration_dict)

    def migrate_from_json(self, filename):
        with open(filename, "r") as f:
            rows = f.readlines()[0].replace("\n", "")
        result = json.loads(rows)
        for dict in result:
            self.process_migration_dict(dict)
        return (result)

    def migration_list_to_json(self, filename=None):
        result = json.dumps(self.migration_list)

        if filename is None or filename == "" or filename.isspace():
            with open("migrations/migration-" + str(self.migration_number) + ".json", "w+") as f:
                f.write(result)
        else:
            with open(f"migrations/{filename}.json", "w+") as f:
                f.write(result)

    def create_migrations_from_df(self, name, dataframe):

        columns, return_types = self.extract_columns_and_types_from_df(dataframe)

        migration_dict = {"create": {"table_name": name, "columns": columns, "types": return_types}}
        self.migration_list.append(migration_dict)
        self.migration_list_to_json()
        # return columns, return_types

    def extract_columns_and_types_from_df(self, dataframe):
        columns = list(dataframe.columns)

        return_types = []

        if columns == []:
            return ["id"], ["int"]

        for column in dataframe:
            if dataframe.empty:
                return_types.append(type(None).__name__)
                continue

            t = dataframe.loc[0, column]
            try:
                if pd.isna(t):
                    return_types.append(type(None).__name__)
                else:
                    try:
                        return_types.append(type(t.item()).__name__)
                    except:
                        return_types.append(type(t).__name__)
            except:
                # length = 2**( int(dataframe[col].str.len().max()) - 1).bit_length()
                length = int(dataframe[column].str.len().max())
                length += 0.1 * length
                length = int(math.ceil(length / 10.0)) * 10
                return_types.append(f'nvarchar({length})' if type(t).__name__ == 'str' else type(t).__name__)

        if (columns[0] != "id"):
            columns.insert(0, "id")
            return_types.insert(0, "int")

        return columns, return_types




import pandas as pd
from typing import Optional, Any
import numpy as np
import abc


class Select:
    def __init__(self, columns:list, table):
        self.columns=columns
        self.table=table
        
    def __str__(self):
        return("SELECT "+",".join(self.columns)+" FROM "+self.table.name)
    
    
class Insert:
    def __init__(self, column_value_dict:dict, table):
        self.column_value_dict=column_value_dict
        self.table=table
    
    def __str__(self):
        return("INSERT INTO "+self.table.name+" ("+",".join(list(self.columns_value_dict.keys()))+") VALUES "+",".join(list(self.columns_value_dict.values()))+")")
        
        
class Update:
    def __init__(self, column_value_dict:dict, table):
        self.column_value_dict=column_value_dict
        self.table=table
    def __str__(self):
        return("UPDATE "+self.table.name+" SET ("+[str(k)+" = "+str(v) for k,v in self.columns_value_dict.items()])
                
        
class Delete:
    def __init__(self, table):
        self.table=table
        
    def __str__(self):
        return("DELETE FROM "+self.table.name)
        
        
class Where:
    def __init__(self, column: str, value: Any, operator="="):
        self.column = column
        self.value = value
        self.operator = operator
    
    def __str__(self):
        return f"WHERE {self.column} {self.operator} {self.value}"
      
class ToDf:
    def __init__(self):
        pass
    
    def __str__(self):
        return("")
    

# Tables
class AbstractSelectable:
    def __init__(self, db1, name, columns=None):
        self.db1 = db1
        self.name = name
        self.columns = columns
        self.query_building_enabled=False
        self._query_blocks=[]
        
        
    def _get_selected_columns(self, query):
        from_keyword_count=query.lower().count("from") #There can be "from" substrings in column names: e.g. SELECT uid,from_node_uid,to_node_uid,channel,visible,pipeline_uid,project_uid FROM edges
        #print("COUNT", from_keyword_count)
        if from_keyword_count==1:
            column_string = query.lower().split("from")[0]
        else:
            column_string = "from".join(query.lower().split("from")[0:(from_keyword_count)])
        #print("column string",column_string)
        if "*" in column_string:
            columns = self.columns
        elif column_string.find(",") == -1:
            assert column_string.count("select")<=1 #assume there are no table columns containing "select" substring
            columns = [column_string.replace("select","").strip()]
        else:
            columns = [x.strip() for x in column_string.split(",")]
        return(columns)


    def _fetch_results(self, columns_count):
        rows = self.db1.cursor.fetchall()
        
        def cast_rows_to_list_of_lists(rows, columns_count):
            cleared_rows = []
            for row in rows:  # Because of unhashable type: 'pyodbc.Row'
                list1 = []
                for i in range(columns_count):
                    list1.append(row[i])
                cleared_rows.append(list1)
            return(cleared_rows)
        
        
        
        results=cast_rows_to_list_of_lists(rows, columns_count)
        
        return (results)


    def select(self, query, flattening_of_results=False):
        """given SELECT query returns Python list"""
        """Columns give the number of selected columns"""
        

        if self.query_building_enabled:
            columns=self._get_selected_columns(query)
            self._query_blocks.append(Select(columns, self))
            return(self) #to enable chaining
        
        else:
            print(query)
            self.db1.cursor.execute(query)
            
            columns_count=len(self._get_selected_columns(query))
            rows=self._fetch_results(columns_count)
                
            def flatten_results(rows, columns_count):
                """Returns flat list for one column and list of lists for multiple columns"""   
                if columns_count == 1:
                    cleared_rows = [item[0] for item in rows]
                return(cleared_rows)
            
            if flattening_of_results:
                rows=flatten_results(rows, columns_count)
                
            return(rows)
            

    def select_all(self):
        all_cols_query = ""
        for col in self.columns:
            all_cols_query = all_cols_query + col + ","
        if all_cols_query[-1] == ",":
            all_cols_query = all_cols_query[:-1]
        list1 = self.select(f"SELECT {all_cols_query} FROM " + self.name)
        return (list1)

    def select_to_df(self):
        rows = self.select_all()
        if self.query_building_enabled:
            self.to_df()
            df=None
        else:
            table_columns = self.columns
            df = pd.DataFrame(rows, columns=table_columns)
        return (df)

    def export_to_xlsx(self, filename="items.xlsx"):
        list1 = self.select_all()
        df1 = pd.DataFrame(list1, columns=["id"] + self.columns)
        df1.to_excel(filename)
        
    def where(self, column, value, operator="="):
        self._query_blocks.append(Where(column, value, operator))
        return(self) #to enable chaining
    
    def to_df(self):
        self._query_blocks.append(ToDf())
        return(self) #to enable chaining
        
    
    def _construct_query_string_from_blocks(self,blocks):
        core_query=" ".join([str(x) for x in blocks if type(x)!=Where])
        where_query=" ".join([str(x) for x in blocks if type(x)==Where])
        where_query=where_query.replace("WHERE","AND")
        where_query=where_query.replace("AND","WHERE",1) #First WHERE is not replaced by AND
        query=" ".join([core_query,where_query])
        print(query)
        return(query)
    
    def _build_queries_from_blocks(self):
        queries=[]
        queries_destinations=[]
        grouped_blocks=[]
        query_destination=None
        
        for i,block in enumerate(self._query_blocks):
            if type(block)!=Where and type(block)!=ToDf:
                grouped_blocks.append([block])
            else:
                grouped_blocks[-1].append(block)
                
            if type(block)==ToDf:
                query_destination="DF"
            
        for i, grouped_block in enumerate(grouped_blocks):
            query=self._construct_query_string_from_blocks(grouped_block)
            queries.append(query)
            queries_destinations.append(query_destination)
            
        assert len(queries) == len(queries_destinations)    
        return(queries, queries_destinations)
    
    def execute(self, query=""):
        """passes execution to DB object"""
        if not self.query_building_enabled:
            self.db1.execute(query)
            
        else:
            queries, queries_destinations=self._build_queries_from_blocks()
            self._query_blocks=[]
            for i,query in enumerate(queries):
                self.db1.execute()
                
                if "SELECT" in query:
                    selected_columns=self._get_selected_columns(query)
                    rows=self._fetch_results(len(selected_columns))
                    
                    
                    if queries_destinations[i]=="DF":
                        df = pd.DataFrame(rows, columns=selected_columns)
                        return(df)
                    else:
                        return(rows)
                    
            

class AbstractJoinable(AbstractSelectable):
    def __init__(self, db1, name, columns=None):
        super().__init__(db1, name, columns)

    def inner_join(self, joinable, column1, column2):
        join_name = self.name + " INNER JOIN " + joinable.name + " ON " + column1 + "=" + column2
        join_columns = list(set(self.columns) | set(joinable.columns))
        new_joinable = AbstractJoinable(self.db1, join_name, join_columns)
        return (new_joinable)



class AbstractTable(AbstractJoinable, abc.ABC):
    def __init__(self, db1, name, columns = None, types: Optional[list[str]] = None, id_column_name: str = "id"):
        """id_column_name -> name of predefined column with autoincrement + PK"""
        super().__init__(db1, name, columns)
        self.types: list[str] = types #[] if types is None else types - is wrong, if types not initialized it should be None, not empty list
        self.id_column_name: str = id_column_name
        assert len(self.columns) == len(self.types)
        self.column_type_dict={self.columns[i]:self.types[i] for i,x in enumerate(self.columns)}

    # Temporary disabled, please make sure this is implemented where needed, don't introduce breaking changes please
    # @abc.abstractmethod
    @classmethod
    def init_from_column_type_dict(cls, db1, name, column_type_dict, id_column_name="id"):
        column_converted_type_dict = db1._convert_column_type_dict_from_python(column_type_dict)
        columns = list(column_converted_type_dict.keys())
        types = list(column_converted_type_dict.values())

        return cls(db1, name, columns, types, id_column_name=id_column_name)
    
    
    def drop(self):
        query = "DROP TABLE " + self.name
        print(query)
        self.execute(query)

    def update(self, variable_assign, where=None):
        if where is None:
            query = "UPDATE " + self.name + " SET " + variable_assign
        else:
            query = "UPDATE " + self.name + " SET " + variable_assign + " WHERE " + where
        print(query)
        return self.execute(query)

    def update_from_df(
            self, update_df: pd.DataFrame, where_column: Optional[str] = None, where_value: Any = None) -> None:
        """Build UPDATE SQL query from a dataframe and execute it.

        :param update_df: Dataframe with updated values - MUST only hold a single row
        :type update_df: pd.DataFrame
        :param where_column: Column name to use for WHERE clause.
        :type where_column: str
        :param where_value: Value to use for WHERE clause.
        :type where_value: Any
        """
        if not len(update_df) == 1:
            raise ValueError("There can only be one row in the UPDATE dataframe")

        types_without_id_column = [type_ for column, type_ in zip(self.columns, self.types) 
                           if column != self.id_column_name]
        if len(types_without_id_column) != len(update_df.columns):
            raise AttributeError(
                "Number of columns in dataframe does not match number of columns in table"
            )

        column_value_string = ""
        for (column, cell_value), column_type in zip(update_df.iloc[0].items(), types_without_id_column):
            if cell_value is None:
                column_value_string += f"{column} = NULL, "
            elif column_type in ["double", "int", "tinyint"]:
                column_value_string += f"{column} = {cell_value}, "
            elif "varchar" in column_type:
                column_value_string += f"{column} = '{cell_value}', "
            elif column_type in ["json", "text", "mediumtext", "longtext", "datetime"]:
                column_value_string += f"{column} = '{cell_value}', "
            else:
                raise AttributeError(f"Unknown column type '{column_type}'")

        column_value_string = column_value_string.rstrip(", ")
        sql_query = f"UPDATE {self.name} SET {column_value_string}"

        if where_column is not None and where_value is not None:
            sql_query += f" WHERE {where_column} = {where_value};"
        else:
            sql_query += ";"

        print(sql_query)
        self.execute(sql_query)

    def _adjust_df(self, df: pd.DataFrame, debug_mode=False) -> pd.DataFrame:
        """
        Adjust DataFrame in case number of its columns differs from number of columns in DB table.
        :param df: original DF
        :param debug_mode: in debug mode a warning will be printed to console
        :return: adjusted DF
        """

        df_copy = df.copy()

        difference_columns_missing = list(set(self.columns).difference(set(df_copy.columns)).difference({'id'}))

        if difference_columns_missing:
            if debug_mode:
                print(f'DataFrame missing columns {difference_columns_missing} will be added as empty.')

            for col in difference_columns_missing:
                df_copy[col] = np.nan

        difference_columns_extra = list(set(df_copy.columns).difference(set(self.columns)))

        if difference_columns_extra:
            if debug_mode:
                print(f'DataFrame extra columns {difference_columns_extra} will be removed.')

            df_copy.drop(difference_columns_extra, axis=1, inplace=True)

        # Ensure the same order of columns in DF as in DB table
        cols_order = [x for x in self.columns if x != 'id']
        df_copy = df_copy[cols_order]

        return df_copy


    def insert_from_df(self, df, batch=1, try_mode=False, debug_mode=False, adjust_df=False, insert_id=False):

        if adjust_df:
            df = self._adjust_df(df, debug_mode)

        if insert_id:
            assert len(df.columns) == len(self.columns)
            assert set(df.columns) == set(self.columns)
            
            inserted_columns=self.columns
            
        else:
            assert len(df.columns) + 1 == len(self.columns) # +1 because of id column
            
            inserted_columns=list(dict.fromkeys(self.columns)) #DEDUPLICATION preserving order -> better than inserted_columns = set(self.columns) 
            id_index=inserted_columns.index(self.id_column_name)
            inserted_columns.pop(id_index)
            print(inserted_columns,df.columns)
            
            assert set(df.columns) == set(inserted_columns) #elements are matchin
            #df = df[inserted_columns]
        df = df[inserted_columns]

        pd_nullable_dtypes = {pd.Int8Dtype(), pd.Int16Dtype(), pd.Int32Dtype(), pd.Int64Dtype(),
                              pd.UInt8Dtype(), pd.UInt16Dtype(), pd.UInt32Dtype(), pd.UInt64Dtype(),
                              pd.Float32Dtype(), pd.Float64Dtype()}
        pd_nullable_columns = df.dtypes.isin(pd_nullable_dtypes)

        # cast pd nullable columns to float - changing to 'NULL' then proceeds without an error.
        # in database floats are again casted to integers where required
        df = df.copy()
        df.loc[:, pd_nullable_columns] = df.loc[:, pd_nullable_columns].astype(float)

        # TODO: handling nan values -> change to NULL
        for column in list(df.columns):
            df.loc[pd.isna(df[column]), column] = "NULL"

        # rows = df.values.tolist()
        # for i, row in enumerate(rows):
        #     for j, record in enumerate(row):
        #         if type(record) == str:
        #             rows[i][j] = "'" + record + "'"
        #print(rows)
        rows = df.values.tolist()
        return self.insert(rows, batch=batch, try_mode=try_mode, debug_mode=False, insert_id=insert_id)

    #TODO: need to solve inserting in different column_order
    #check df column names, permute if needed
    def insert_from_column_value_dict(self, dict, insert_id=False):
        df = pd.DataFrame(dict, index=[0])
        return self.insert_from_df(df, insert_id=insert_id)

    def insert_from_column_value_dict_list(self, list, insert_id=False):
        df = pd.DataFrame(list)
        self.insert_from_df(df, insert_id=insert_id)


    def delete(self, where=None):

        if where is None:
            query = "DELETE FROM " + self.name
        else:
            query = "DELETE FROM " + self.name + " WHERE " + where
        return self.execute(query)




    


import pandas as pd


def operation_sequence1(testing_table, *args):
    testing_table.delete(args[0])
    testing_table.delete(args[1])
    testing_table.update(args[2], args[3])

input_df = pd.DataFrame(
    {"name": ["John", "Lisa", "Charlie", "Rachel"], "age": [5, 67, 22, 144]}
).astype("object")

compare_df = pd.DataFrame({"name": ["Charlie", "Rachel"], "age": [22, 100]}).astype(
    "object"
)

test_case1 = (input_df, compare_df, operation_sequence1)


#TODO: add more test cases


import pytest
import dbhydra.dbhydra_core as dh
import pandas as pd
import keepvariable.keepvariable_core as kv

import dbhydra.tests.test_cases as test_cases

db_details_mysql = kv.load_variable_safe("tests/test_sql_db.kv", varname="db_details")
db_details_postgres = kv.load_variable_safe(
    "tests/test_postgres_db.kpv", varname="db_details"
)


@pytest.fixture(scope="module")
def mysqldb():
    return dh.Mysqldb(db_details=db_details_mysql)


@pytest.fixture(scope="module")
def postgresdb():
    return dh.PostgresDb(db_details=db_details_postgres)


@pytest.mark.parametrize("db", [("mysqldb"), ("postgresdb")])
def test_mysql_server_connection_and_table_presence(db, request):
    db = request.getfixturevalue(db)
    with db.connect_to_db():
        tables = db.generate_table_dict()

    assert "testing_table" in tables


@pytest.mark.parametrize(
    "db,table_name,compare_df",
    [
        (
            "mysqldb",
            "testing_table",
            pd.DataFrame(
                {
                    "id": [1, 2, 3],
                    "value1": [100, 100, 100],
                    "value2": ["slon", "slon", "slon"],
                }
            ).astype("object"),
        ),
        (
            "mysqldb",
            "testing_table_insert",
            pd.DataFrame([], columns=["id", "name", "age"]).astype("object"),
        ),
        (
            "postgresdb",
            "testing_table",
            pd.DataFrame(
                {
                    "id": [1, 2, 3],
                    "value1": [100, 100, 100],
                    "value2": ["slon", "slon", "slon"],
                }
            ).astype("object"),
        ),
        (
            "postgresdb",
            "testing_table_insert",
            pd.DataFrame([], columns=["id", "name", "age"]).astype("object"),
        ),
    ],
)
def test_mysql_select(db, table_name, compare_df, request):
    db = request.getfixturevalue(db)
    with db.connect_to_db():
        tables = db.generate_table_dict()
        testing_table = tables[table_name]
        response_df = testing_table.select_to_df().astype("object")

    assert response_df.equals(compare_df.reindex(columns=response_df.columns))



@pytest.mark.parametrize(
    "db,input_df,compare_df,seq,",
    [
        ("mysqldb", test_cases.test_case1[0], test_cases.test_case1[1], test_cases.test_case1[2]),
        ("postgresdb", test_cases.test_case1[0], test_cases.test_case1[1], test_cases.test_case1[2]),
    ],
)
def test_mysql_crud(db, input_df, compare_df, seq, request):
    db = request.getfixturevalue(db)

    with db.connect_to_db():
        tables = db.generate_table_dict()
        testing_table = tables["testing_table_insert"]
        testing_table.insert_from_df(input_df)
        seq(testing_table, "age=5", "name='Lisa'", "age=100", "age=144")
        response_df = testing_table.select_to_df()
        testing_table.delete()

    response_rows = response_df.drop(["id"], axis=1).astype("object")

    assert response_rows.equals(compare_df)



import sys

from dbhydra.src.abstract_db import AbstractDb
from dbhydra.src.tables import SqlServerTable
from dbhydra.src.errors.exceptions import DbHydraException


#! Disabled on macOS --> problematic import
if sys.platform != "darwin":
    import pyodbc



class SqlServerDb(AbstractDb):
    matching_table_class = SqlServerTable
    
    def connect_remotely(self):
        if sys.platform == "darwin":
            raise DbHydraException("pyodbc library (MSSQL) not supported on macOS.")

        self.connection = pyodbc.connect(
            r'DRIVER={' + self.DB_DRIVER + '};'
            r'SERVER=' + self.DB_SERVER + ';'
            r'DATABASE=' + self.DB_DATABASE + ';'
            r'UID=' + self.DB_USERNAME + ';'
            r'PWD=' + self.DB_PASSWORD + '',
            timeout=1
        )
        self.cursor = self.connection.cursor()
        print("DB connection established")

    def connect_locally(self):
        if sys.platform == "darwin":
            raise DbHydraException("pyodbc library (MSSQL) not supported on macOS.")
        
        self.connection = pyodbc.connect(
            r'DRIVER={' + self.DB_DRIVER + '};'
            r'SERVER=' + self.DB_SERVER + ';'
            r'DATABASE=' + self.DB_DATABASE + ';'
            r'TRUSTED_CONNECTION=yes;',
            timeout=1
            # r'PWD=' + self.DB_PASSWORD + '')
        )

        self.cursor = self.connection.cursor()
        print("DB connection established")

    def get_all_tables(self):
        sysobjects_table = SqlServerTable(self, "sysobjects", ["name"], ["nvarchar(100)"])
        query = "select name from sysobjects where xtype='U'"
        rows = sysobjects_table.select(query)
        return (rows)

    def generate_table_dict(self):
        tables = self.get_all_tables()
        table_dict = dict()
        for i, table in enumerate(tables):
            table_dict[table] = SqlServerTable.init_all_columns(self, table)

        return (table_dict)

    def get_foreign_keys_columns(self):
        sys_foreign_keys_columns_table = SqlServerTable(self, "sys.foreign_key_columns",
                                               ["parent_object_id", "parent_column_id", "referenced_object_id",
                                                "referenced_column_id"], ["int", "int", "int", "int"])
        query = "select parent_object_id,parent_column_id,referenced_object_id,referenced_column_id from sys.foreign_key_columns"
        rows = sys_foreign_keys_columns_table.select(query)

        sys_foreign_keys_columns_table = SqlServerTable(self, "sys.tables", ["object_id", "name"], ["int", "nvarchar(100)"])
        query = "select object_id,name from sys.tables"
        table_names = sys_foreign_keys_columns_table.select(query)

        table_id_name_dict = {x[0]: x[1] for x in table_names}

        foreign_keys = []
        for i, row in enumerate(rows):
            fk = {"parent_table": table_id_name_dict[row[0]], "parent_column_id": row[1] - 1,
                  "referenced_table": table_id_name_dict[row[2]],
                  "referenced_column_id": row[3] - 1}  # minus 1 because of indexing from 0
            foreign_keys.append(fk)

        return (foreign_keys)



class db(SqlServerDb):
    """Deprecated - do not remove until dbhydra 3.x"""
    def __init__(self, config_file="config.ini", db_details=None):
        print("Deprecation warning!, db was renamed to SQLServerDb and the old name will deprecated in future!")
        super().__init__(config_file=config_file, db_details=db_details)





import pymysql
from dbhydra.src.abstract_db import AbstractDb
from dbhydra.src.tables import MysqlTable



class MysqlDb(AbstractDb):
    matching_table_class = MysqlTable
    python_database_type_mapping = PYTHON_TO_MYSQL_DATA_MAPPING = \
    {
    'int': "int",
    'float': "double",
    'str': "nvarchar(2047)",
    'tuple': "json",
    'list': "nvarchar(2047)",
    'dict': "nvarchar(2047)",
    'bool': "tinyint",
    'datetime': "datetime",
    'Jsonable': "json"
    }

    def connect_locally(self):
        self.connection = pymysql.connect(host=self.DB_SERVER, user=self.DB_USERNAME, password=self.DB_PASSWORD,
                                          database=self.DB_DATABASE)
        self.cursor = self.connection.cursor()
        print("DB connection established")

    def connect_remotely(self):
        if self.DB_PORT is not None:
            self.connection = pymysql.connect(host=self.DB_SERVER, port=self.DB_PORT, user=self.DB_USERNAME,
                                              password=self.DB_PASSWORD, database=self.DB_DATABASE)
        else:
            self.connection = pymysql.connect(host=self.DB_SERVER, user=self.DB_USERNAME, password=self.DB_PASSWORD,
                                              database=self.DB_DATABASE)
        self.cursor = self.connection.cursor()
        print("DB connection established")

    def create_new_db(self):
        create_db_command = "CREATE DATABASE " + self.DB_DATABASE
        self.execute(create_db_command)

        
    def execute(self, query, is_autocommitting=True):
        result=self.cursor.execute(query)
        if is_autocommitting:
            self.connection.commit()
        return result
    

    def get_all_tables(self):
        sysobjects_table = MysqlTable(self, "information_schema.tables", ["TABLE_NAME"], ["nvarchar(100)"])
        query = "SELECT TABLE_NAME,TABLE_TYPE,TABLE_SCHEMA FROM information_schema.tables where TABLE_TYPE='BASE TABLE' AND TABLE_SCHEMA='" + self.DB_DATABASE + "' ;"
        rows = sysobjects_table.select(query)
        tables = [x[0] for x in rows]
        return (tables)

    def generate_table_dict(self):
        tables = self.get_all_tables()
        table_dict = dict()
        for i, table in enumerate(tables):
            table_dict[table] = MysqlTable.init_all_columns(self, table)
        return (table_dict)



class Mysqldb(MysqlDb):
    """Deprecated - do not remove until dbhydra 3.x"""
    def __init__(self, config_file="config.ini", db_details=None):
        print("Deprecation warning!, Mysqldb was renamed to MysqlDb and the old name will deprecated in future!")
        super().__init__(config_file=config_file, db_details=db_details)



from dbhydra.src.abstract_db import AbstractDb
from dbhydra.src.tables import PostgresTable

from sys import platform
if platform != "linux" and platform != "linux2":
    # linux
    import psycopg2 # disable dependency for server (Temporary)


class PostgresDb(AbstractDb):
    matching_table_class = PostgresTable

    def connect_locally(self):
        self.connection = psycopg2.connect(
            host=self.DB_SERVER,
            database=self.DB_DATABASE,
            user=self.DB_USERNAME,
            password=self.DB_PASSWORD)
        self.cursor = self.connection.cursor()

    def connect_remotely(self):
        self.connection = psycopg2.connect(
            host=self.DB_SERVER,
            database=self.DB_DATABASE,
            user=self.DB_USERNAME,
            password=self.DB_PASSWORD)
        self.connection.autocommit = True
        self.cursor = self.connection.cursor()
        
    def close_connection(self):
        self.cursor.close()
        print("DB connection closed")

    def execute(self, query, is_autocommitting=True):
        result=self.cursor.execute(query)
        if is_autocommitting:
            self.connection.commit()
        return result

        # return  [''.join(i) for i in self.cursor.fetchall()]

    def get_all_tables(self):
        self.cursor.execute("""SELECT table_name FROM information_schema.tables
               WHERE table_schema = 'public'""")
        tables = [''.join(x) for x in self.cursor.fetchall()]
        return tables

    def generate_table_dict(self):
        tables = self.get_all_tables()
        table_dict = dict()
        for i, table in enumerate(tables):
            table_dict[table] = PostgresTable.init_all_columns(self, table)

        return (table_dict)



import abc
import threading
from contextlib import contextmanager

from dbhydra.src.migrator import Migrator
from dbhydra.src.tables import AbstractTable

def read_connection_details(config_file):
    def read_file(file):
        """Reads txt file -> list"""
        with open(file, "r") as f:
            rows = f.readlines()
            for i, row in enumerate(rows):
                rows[i] = row.replace("\n", "")
        return (rows)
    
    connection_details = read_file(config_file)
    db_details = {}
    for detail in connection_details:
        key = detail.split("=")[0]
        value = detail.split("=")[1]
        db_details[key] = value

        # Skip empty lines to avoid error when reading config file
        if not detail:
            continue

    print(", ".join([db_details['DB_SERVER'], db_details['DB_DATABASE'], db_details['DB_USERNAME']]))
    return (db_details)





class Transaction:
    def __init__(self, connection):
        self.connection = connection

    def commit(self):
        self.connection.commit()

    def rollback(self):
        self.connection.rollback()




class AbstractDb(abc.ABC):
    matching_table_class = AbstractTable
    
    typing_to_python_mapping = {
        'List': 'list',
        'Dict': 'dict',
        'Tuple': 'tuple',
        'Set': 'set',
        'Union': 'object',
        'Optional': 'object',
        'Jsonable': 'Jsonable',
        # 'FrozenSet': frozenset,
        # 'Deque': list,
        # 'Any': object,
        #
        # 'Callable': object,
        # 'Type': type,
        # 'TypeVar': object,
        # 'Generic': object,
        # 'Sequence': list,
        # 'Iterable': list,
        # 'Mapping': dict,
        # 'AbstractSet': set,
    }
    python_database_type_mapping = {}
    def __init__(self, config_file="config.ini", db_details=None):
        if db_details is None:
            db_details = read_connection_details(config_file)

        self.locally = True
        if db_details["LOCALLY"] == "False":
            self.locally = False

        self.DB_SERVER = db_details["DB_SERVER"]
        self.DB_DATABASE = db_details["DB_DATABASE"]
        self.DB_USERNAME = db_details["DB_USERNAME"]
        self.DB_PASSWORD = db_details["DB_PASSWORD"]
        
        self.is_autocommiting=True

        if "DB_PORT" in db_details.keys():
            self.DB_PORT = int(db_details["DB_PORT"])
        else:
            self.DB_PORT = None

        if "DB_DRIVER" in db_details.keys():
            self.DB_DRIVER = db_details["DB_DRIVER"]
        else:
            self.DB_DRIVER = "ODBC Driver 13 for SQL Server"

        self.lock = threading.Lock()

        self.connect_to_db()

    @abc.abstractmethod
    def connect_locally(self):
        pass

    @abc.abstractmethod
    def connect_remotely(self):
        pass

    def _connect(self):
        if self.locally:
            self.connect_locally()
        else:
            self.connect_remotely()

    def connect(self):
        print("DEPRECATION WARNING: use `connect_to_db` context manager instead of `connect` method")
        if self.locally:
            self.connect_locally()
        else:
            self.connect_remotely()

    #@abc.abstractmethod
    def get_all_tables(self):
        pass

    #@abc.abstractmethod
    def generate_table_dict(self):
        pass

    def get_table(self, table_name: str):
        """
        Retrieves Table from DB using its name.
        """

        try:
            table = self.generate_table_dict()[table_name]
        except KeyError:
            print(f'Table "{table_name}" was not found in the DB.')
            raise KeyError

        return table

    @contextmanager
    def connect_to_db(self):
        try:
            self._connect()
            yield None
        finally:
            self.close_connection()

    @contextmanager
    def connect_to_table(self, table_name):
        with self.connect_to_db():
            table = self.generate_table_dict()[table_name]

            yield table


    @contextmanager
    def transaction(self):
        self.is_autocommiting=False
        transaction = Transaction(self.connection)
        self.active_transactions.append(transaction)
        try:
            yield transaction
        except Exception as e:
            transaction.rollback()
            raise e
        else:
            transaction.commit()
        finally:
            self.active_transactions.remove(transaction)
            if len(self.active_transactions)==0:
                self.is_autocommiting=True


    def execute(self, query):
        result=self.cursor.execute(query)
        if self.is_autocommitting:
            self.cursor.commit()
        return(result)

    def close_connection(self):
        self.connection.close()
        print("DB connection closed")

    def initialize_migrator(self):
        self.migrator = Migrator(self)

    def _convert_column_type_dict_from_python(self, column_type_dict):
        """
        First apply mapping from python typing module to standard python.
        Then apply mapping from python to database types
        """
        typing_python_mapping = {column_name: self.typing_to_python_mapping.get(column_type,column_type) for column_name, column_type in column_type_dict.items()}
        return {column_name: self.python_database_type_mapping[column_type] for column_name, column_type in typing_python_mapping.items()}




"""DB Hydra ORM"""
import abc
from typing import Any

import pandas as pd
from pydantic import BaseModel

##### Do not remove imports - they are expored in the package
from dbhydra.src.sqlserver_db import SqlServerDb, db
from dbhydra.src.mysql_db import MysqlDb, Mysqldb
from dbhydra.src.bigquery_db import BigQueryDb
from dbhydra.src.mongo_db import MongoDb
from dbhydra.src.postgres_db import PostgresDb
from dbhydra.src.xlsx_db import XlsxDb, XlsxDB
from dbhydra.src.abstract_db import AbstractDb
from dbhydra.src.tables import SqlServerTable, PostgresTable, MysqlTable, XlsxTable, AbstractTable, MongoTable, BigQueryTable, Table, AbstractSelectable, AbstractJoinable
##### Do not remove imports - they are expored in the package


class Jsonable(str):
    """Is used as type in python_database_type_mapping"""    
    pass



# dataframe - dictionary auxiliary functions
def df_to_dict(df, column1, column2):
    dictionary = df.set_index(column1).to_dict()[column2]
    return (dictionary)


def dict_to_df(dictionary, column1, column2):
    df = pd.DataFrame(list(dictionary.items()), columns=[column1, column2])
    return (df)



from pydantic_core import CoreSchema, core_schema
from pydantic import GetCoreSchemaHandler


class AbstractModel(abc.ABC, BaseModel):
    @classmethod
    def generate_dbhydra_table(cls, table_class: AbstractTable, db1, name, id_column_name="id"):
        column_type_dict = create_table_structure_dict(cls, id_column_name=id_column_name)
        dbhydra_table = table_class.init_from_column_type_dict(db1, name, column_type_dict, id_column_name=id_column_name)
        return dbhydra_table
    
    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

 

def create_table_structure_dict(api_class_instance, id_column_name="id"):
    """
    Accepts instance of API data class (e.g. APIDatabase) and converts it to dictionary {attribute_name: attribute_type}
    """
    table_structure_dict = {id_column_name: "int"}
    table_structure_dict = {**table_structure_dict,
                            **{attribute_name: attribute_type.__name__ for attribute_name, attribute_type in api_class_instance.__annotations__.items()}}

    return table_structure_dict



import pandas as pd
import numpy as np
from typing import Optional, Any
import abc

#xlsx imports
import pathlib

from dbhydra.src.abstract_table import AbstractTable, AbstractSelectable, AbstractJoinable


MONGO_OPERATOR_DICT = {"=": "$eq", ">": "$gt", ">=": "$gte", " IN ": "$in", "<": "$lt", "<=": "$lte", "<>": "$ne"}

POSTGRES_TO_MYSQL_DATA_MAPPING = {
    "int": "int",
    "integer": "int",
    "bigint": "bigint",
    "smallint": "smallint",
    "character varying": "varchar",
    "text": "longtext",
    "boolean": "tinyint",
    "double precision": "double",
    "real": "float",
    "numeric": "decimal",
    "date": "date",
    "timestamp": "timestamp"
}

PYTHON_TO_MYSQL_DATA_MAPPING = {
    'int': "int",
    'float': "double",
    'str': "nvarchar(2047)",
    'bool': "bit",
    'datetime': "datetime"
}




def save_migration(function, *args, **kw):  # decorator
    def new_function(instance, *args, **kw):
        print("TOTO TU")
        print(instance)
        print(*args)
        command = function.__name__
        if command == "create":
            migration_dict = {
                "create": {"table_name": instance.name, "columns": instance.columns, "types": instance.types}}
            print(migration_dict)
        if command == "drop":
            migration_dict = {"drop": {"table_name": instance.name}}
            print(migration_dict)
        if command == "add_column":
            migration_dict = {
                "add_column": {"table_name": instance.name, "column_name": args[0], "column_type": args[1]}}
            print(migration_dict)
        if command == "drop_column":
            migration_dict = {
                "drop_column": {"table_name": instance.name, "column_name": args[0]}}
            print(migration_dict)
        if command == "modify_column":
            migration_dict = {
                "modify_column": {"table_name": instance.name, "column_name": args[0], "column_type": args[1]}}
            print(migration_dict)
        # TODO: add other methods

        migrator = instance.db1.migrator
        migrator.migration_list.append(migration_dict)
        # migrator.migration_list_to_json()
        function(instance, *args, **kw)

    return (new_function)



######### POSTGRES ##############

class PostgresTable(AbstractTable):
    def __init__(self, db1, name, columns=None, types=None):
        super().__init__(db1, name, columns)
        self.types = types

        print("==========================================")

    def initialize_columns(self):
        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "';"
        columns = information_schema_table.select(query)
        self.columns = columns

    def initialize_types(self):
        self.types = self.get_all_types()

    def get_all_columns(self):
        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
        query = "SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
        columns = information_schema_table.select(query)

        return (columns)

    def convert_types_from_mysql(self):
        inverse_dict_mysql_to_postgres = dict(zip(POSTGRES_TO_MYSQL_DATA_MAPPING.values(), POSTGRES_TO_MYSQL_DATA_MAPPING.keys()))
        postgres_types = list(map(lambda x: inverse_dict_mysql_to_postgres.get(x, x), self.types))
        self.types = postgres_types

    def get_all_types(self):

        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
        query = "SELECT DATA_TYPE,character_maximum_length FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
        types = information_schema_table.select(query)
        data_types = [x[0].lower() for x in types]
        date_lengths = [x[1] for x in types]

        mysql_types = list(map(lambda x: POSTGRES_TO_MYSQL_DATA_MAPPING.get(x, x), data_types))


        for i in range(len(mysql_types)):
            if date_lengths[i] is not None:
                mysql_types[i] = mysql_types[i] + f"({date_lengths[i]})"

        return (mysql_types)


    def select_all(self):
        print(super().select_all())
        return [i for i in super().select_all()]

    @classmethod
    def init_all_columns(cls, db1, name):
        temporary_table = cls(db1, name)
        columns = temporary_table.get_all_columns()
        types = temporary_table.get_all_types()

        if temporary_table.id_column_name in columns:
            id_col_index = columns.index(temporary_table.id_column_name)
            columns.pop(id_col_index)
            columns.insert(0, temporary_table.id_column_name)
            types.pop(id_col_index)
            types.insert(0, "int")

        return (cls(db1, name, columns, types))

    # @save_migration
    def create(self, foreign_keys=None):
        assert len(self.columns) == len(self.types)
        assert self.columns[0] == self.id_column_name
        assert self.types[0].lower() == "int" or self.types[0].lower() == "integer"
        query = "CREATE TABLE " + self.name + "("+self.id_column_name+" SERIAL PRIMARY KEY,"
        for i in range(1, len(self.columns)):
            query += self.columns[i] + " " + self.types[i] + ","

        query = query[:-1]
        query += ");"
        print(query)
        try:
            self.db1.execute(query)
        except Exception as e:
            print("Table " + self.name + " already exists:", e)
            print("Check the specification of table columns and their types")

    def insert(self, rows, batch=1, replace_apostrophes=True, try_mode=False, debug_mode=False, insert_id=False):
        start_index = 0 if insert_id else 1
        assert len(self.columns) == len(self.types)
        for k in range(len(rows)):
            if k % batch == 0:
                query = "INSERT INTO " + self.name + " ("
                for i in range(start_index, len(self.columns)):
                    if i < len(rows[k]) + 1:
                        query += self.columns[i] + ","
                if len(rows) < len(self.columns):
                    print(len(self.columns) - len(rows), "columns were not specified")
                if query[-1] == ',':
                    query = query[:-1]
                    query = query + ") VALUES "
                elif query[-1] == '(':
                    query = query[:-1]
                    query = query + " VALUES "



            query += "("
            for j in range(len(rows[k])):
                if rows[k][j] == "NULL" or rows[k][j] == "'NULL'" or rows[k][j] == None or rows[k][
                    j] == "None":  # NaN hodnoty
                    if "int" in self.types[j + start_index]:

                        if replace_apostrophes:
                            rows[k][j] = str(rows[k][j]).replace("'", "")
                        query += "NULL,"
                    else:
                        query += "NULL,"
                elif "nvarchar" in self.types[j + start_index]:
                    if replace_apostrophes:
                        rows[k][j] = str(rows[k][j]).replace("'", "")
                    query += "N'" + str(rows[k][j]) + "',"
                elif "varchar" in self.types[j + start_index]:
                    if replace_apostrophes:
                        rows[k][j] = str(rows[k][j]).replace("'", "")
                    query += "'" + str(rows[k][j]) + "',"
                elif self.types[j + start_index] == "int":
                    query += str(rows[k][j]) + ","
                elif "datetime" in self.types[j + start_index]:
                    if replace_apostrophes:
                        rows[k][j] = str(rows[k][j]).replace("'", "")
                    query += "'" + str(rows[k][j]) + "',"
                elif "date" in self.types[j + start_index]:
                    query += "'" + str(rows[k][j]) + "',"



                else:
                    query += str(rows[k][j]) + ","
            if query[-1] == ",":
                query = query[:-1]
            elif query[-1] == '(':
                query = query + "DEFAULT"

            query = query + "),"
            if k % batch == batch - 1 or k == len(rows) - 1:
                query = query[:-1]

                if debug_mode:
                    print(query)

                if not try_mode:
                    return self.db1.execute(query)
                else:
                    try:
                        return self.db1.execute(query)
                    except Exception as e:

                        print("Query", query, "Could not be inserted:", e)

                        # Write to logs only in debug mode
                        if debug_mode:
                            with open("log.txt", "a") as file:
                                file.write("Query " + str(query) + " could not be inserted:" + str(e) + "\n")

    @save_migration
    def add_column(self, column_name, column_type):
        assert len(column_name) > 1
        command = "ALTER TABLE " + self.name + " ADD COLUMN " + column_name + " " + column_type
        try:
            self.db1.execute(command)
            self.columns.append(column_name)
            self.types.append(column_type)
        except Exception as e:
            print(e)
            print("Cant add column to table.")

    @save_migration
    def drop_column(self, column_name):
        assert len(column_name) > 1
        command = "ALTER TABLE " + self.name + " DROP COLUMN " + column_name
        try:
            self.db1.execute(command)
            index = self.db1.columns.index(column_name)
            self.db1.columns.remove(column_name)
            self.db1.types.remove(self.db1.types[index])
        except Exception as e:
            print(e)
            print("Cant drop " + self.name)

    @save_migration
    def modify_column(self, column_name, new_column_type):
        assert len(column_name) > 1
        command = "ALTER TABLE " + self.name + " ALTER COLUMN " + column_name + " TYPE " + new_column_type
        try:
            self.db1.execute(command)
            index = self.columns.index(column_name)
            self.types[index] = new_column_type
        except Exception as e:
            print("Cant add column to table.")

class BigQueryTable(AbstractSelectable):
    def __init__(self, db1, name, columns=None, types=None):
        self.name = name
        self.db1 = db1
        self.columns = columns
        self.types = types

    @classmethod
    def init_all_columns(cls, db1, name):
        temporary_table = cls(db1, name)
        columns,types = temporary_table.get_all_columns_and_types()

        return (cls(db1, name, columns, types))

    def get_all_columns_and_types(self):
        results = self.db1.client.get_table(f"{self.db1.credentials.project_id}.{self.db1.dataset}.{self.name}")
        column_names = [x.name for x in results.schema ]
        column_types = [x.field_type for x in results.schema]
        return column_names, column_types


    def select(self, query):

        """given SELECT query returns Python list"""
        """Columns give the number of selected columns"""
        print(query)
        # rows =  self.db1.client.query(query).result()
        rows = self.db1.execute(query)
        return rows


############ MONGO ############

class MongoTable():
    def __init__(self, db, name, columns=[], types=[]):
        self.name = name
        self.db1 = db
        print("==========================================")
        print(type(db))
        print(db)
        self.columns = columns
        self.types = types
        self.collection = self.db1.create_table(name)

    def create(self):
        pass

    def drop(self):
        return self.collection.drop()

    def update_collection(self):
        self.collection = self.db1.create_table(self.name)

    def insert(self, document):
        return self.collection.insert_one(document)

    def insert_more(self, documents):
        return self.collection.insert_many(documents)

    def insertMore(self, documents):
        print("WARNING: `insertMore` method will be deprecated in favor of `insert_more`")
        return self.insert_more(documents)

    def select(self, query, columns={}):

        if columns == '*':
            columns = {}

        if (len(columns) == 0):

            return list(self.collection.find(query))
        else:
            return list(self.collection.find(query, columns))

    def select_all(self, query={}):
        return list(self.collection.find(query))

    def select_sort(self, query, fieldname, direction, columns={}):
        if (len(columns) == 0):
            return list(self.collection.find(query).sort(fieldname, direction))
        else:
            return list(self.collection.find(query, columns).sort(fieldname, direction))

    def selectSort(self, query, fieldname, direction, columns={}):
        print("WARNING: `selectSort` method will be deprecated in favor of `select_sort`")
        return self.select_sort(query, fieldname, direction, columns)

    def delete(self, query={}):
        self.collection = self.db1.create_table(self.name)
        return self.collection.delete_many(query)

    def update(self, newvalues, query):
        return self.collection.update_many(query, newvalues)

    def insert_from_df(self, dataframe, insert_id=None):
        if dataframe.empty:
            return
        dataframe = dataframe.replace({pd.NA: None})
        dict_from_df = dataframe.to_dict('records')
        # dict_from_df = dataframe.apply(lambda x : x.dropna().to_dict(),axis=1).tolist() #get rid of nans
        return self.collection.insert_many(dict_from_df)

    def insertFromDataFrame(self, dataframe):
        print("WARNING: `insertFromDataFrame` method will be deprecated in favor of `insert_from_df`")
        return self.insert_from_df(dataframe)

    def select_to_df(self, query={}):
        print(type(pd.DataFrame(list(self.collection.find(query)))))
        print(pd.DataFrame(list(self.collection.find(query))))
        return pd.DataFrame(list(self.collection.find(query)))

    @classmethod
    def init_all_columns(cls, db1, name):

        temporary_table = cls(db1, name)
        values = temporary_table.get_columns_types()
        columns = values[0][1:]
        types = values[1][1:]

        if temporary_table.id_column_name in columns:
            index = columns.index(temporary_table.id_column_name)
            columns.pop(index)
            types.pop(index)

        columns.insert(0, temporary_table.id_column_name)
        types.insert(0, "int")
        types = [x.lower() for x in types]
        types_ = [PYTHON_TO_MYSQL_DATA_MAPPING[x] for x in types]
        types = types_
        return (cls(db1, name, columns, types))

    def keys_exists(self, element, *keys):
        '''
        Check if *keys (nested) exists in `element` (dict).
        '''
        if not isinstance(element, dict):
            raise AttributeError('keys_exists() expects dict as first argument.')
        if len(keys) == 0:
            raise AttributeError('keys_exists() expects at least two arguments, one given.')

        _element = element
        for key in keys:
            try:
                _element = _element[key]
            except KeyError:
                return False
        return True

    def print_nested_keys(self, d, columns, types, parent=""):

        for k in d.keys():

            typ = type(d.get(k)).__name__

            if self.keys_exists(types, k, typ):
                types[k][typ] = types[k][typ] + 1
            else:
                try:
                    types[k][typ] = 1
                except:
                    types[k] = {}
                    types[k][typ] = 1
            if parent + k not in columns:
                columns.append(parent + k)
            if type(d[k]) == dict:
                self.print_nested_keys(d[k], columns, types, parent + k + ".")

    def get_columns_types(self):
        columns = []
        types = {}
        for dict_j in self.collection.find():
            self.print_nested_keys(dict_j, columns, types)
        types = self.get_all_types(types)
        self.columns = columns
        self.types = types
        return columns, types

    def get_all_types(self, types):
        print(types)
        types_list = []
        for k in types.keys():
            values = types.get(k)
            if (len(values) == 1):

                types_list.append(next(iter(values)))
            else:
                chosen = ""
                chosen_cnt = 0
                for t in values.keys():
                    if values.get(t) > chosen_cnt:
                        chosen = t
                types_list.append(chosen + " ?")
        print(types_list)
        return types_list


############## SQL SERVER ##########################


class SqlServerTable(AbstractTable):
    def __init__(self, db1, name, columns=None, types=None):
        """Override joinable init"""
        super().__init__(db1, name, columns)
        self.types = types

    @classmethod
    def init_all_columns(cls, db1, name):
        temporary_table = cls(db1, name)
        columns = temporary_table.get_all_columns()
        types = temporary_table.get_all_types()
        return (cls(db1, name, columns, types))

    def get_all_columns(self):
        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
        query = "SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
        columns = information_schema_table.select(query)
        return (columns)

    def get_all_types(self):
        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
        query = "SELECT DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
        types = information_schema_table.select(query)
        return (types)

    def create(self):
        assert len(self.columns) == len(self.types)
        assert self.columns[0] == self.id_column_name
        assert self.types[0].lower() == "int"
        query = "CREATE TABLE " + self.name + "("+self.id_column_name+" INT IDENTITY(1,1) NOT NULL,"
        for i in range(1, len(self.columns)):
            query += self.columns[i] + " " + self.types[i] + ","
        query += "PRIMARY KEY("+self.id_column_name+"))"

        print(query)
        try:
            self.db1.execute(query)
        except Exception as e:
            print("Table " + self.name + " already exists:", e)
            print("Check the specification of table columns and their types")

    def insert(self, rows, batch=1, replace_apostrophes=True, try_mode=False, debug_mode=False):

        assert len(self.columns) == len(self.types)
        for k in range(len(rows)):
            if k % batch == 0:
                query = "INSERT INTO " + self.name + " ("
                for i in range(1, len(self.columns)):
                    if i < len(rows[k]) + 1:
                        query += self.columns[i] + ","
                if len(rows) < len(self.columns):
                    print(len(self.columns) - len(rows), "columns were not specified")
                query = query[:-1] + ") VALUES "

            query += "("

            for j in range(len(rows[k])):
                if rows[k][j] == "NULL" or rows[k][j] == "'NULL'" or rows[k][j] == None or rows[k][
                    j] == "None":  # NaN hodnoty
                    query += "NULL,"
                elif "nvarchar" in self.types[j + 1]:
                    if replace_apostrophes:
                        rows[k][j] = str(rows[k][j]).replace("'", "")
                    query += "N'" + str(rows[k][j]) + "',"
                elif "varchar" in self.types[j + 1]:
                    if replace_apostrophes:
                        rows[k][j] = str(rows[k][j]).replace("'", "")
                    query += "'" + str(rows[k][j]) + "',"
                elif self.types[j + 1] == "int":
                    query += str(rows[k][j]) + ","
                elif "date" in self.types[j + 1]:
                    query += "'" + str(rows[k][j]) + "',"
                elif "datetime" in self.types[j + 1]:
                    query += "'" + str(rows[k][j]) + "',"
                else:
                    query += str(rows[k][j]) + ","

            query = query[:-1] + "),"
            if k % batch == batch - 1 or k == len(rows) - 1:
                query = query[:-1]

                if debug_mode:
                    print(query)

                if not try_mode:
                    self.db1.execute(query)
                else:
                    try:
                        self.db1.execute(query)
                    except Exception as e:

                        print("Query", query, "Could not be inserted:", e)

                        # Write to logs only in debug mode
                        if debug_mode:
                            with open("log.txt", "a") as file:
                                file.write("Query " + str(query) + " could not be inserted:" + str(e) + "\n")

    def get_foreign_keys_for_table(self, table_dict, foreign_keys):
        # table_dict is in format from db function: generate_table_dict()
        # foreign_keys are in format from db function: get_foreign_keys_columns()
        parent_foreign_keys = []
        for i, fk in enumerate(foreign_keys):
            if fk["parent_table"] == self.name:

                try:
                    print(fk["parent_column_id"])
                    print(self.columns)
                    print(self.name)
                    fk["parent_column_name"] = self.columns[fk["parent_column_id"]]
                    fk["referenced_column_name"] = table_dict[fk["referenced_table"]].columns[
                        fk["referenced_column_id"]]
                    parent_foreign_keys.append(fk)
                except IndexError as e:
                    print("Warning: IndexError for foreign key self.columns[fk[parent_column_id]]:", e)
        return (parent_foreign_keys)


class Table(SqlServerTable):
    """Deprecated - do not remove until dbhydra 3.x"""
    def __init__(self, config_file="config.ini", db_details=None):
        print("Deprecation warning!, Table was renamed to SqlServerTable and the old name will deprecated in future!")
        super().__init__(config_file=config_file, db_details=db_details)



########### MYSQL #############



class MysqlTable(AbstractTable):
    def __init__(self, db1, name, columns=None, types=None, id_column_name = "id"):
        super().__init__(db1, name, columns, types)
        self.id_column_name = id_column_name

    #Disabled because it is inherited
    # @classmethod
    # def init_from_column_type_dict(cls, db1, name, column_type_dict, id_column_name="id"):
    #     column_converted_type_dict = db1._convert_column_type_dict_from_python(column_type_dict)
    #     columns = list(column_converted_type_dict.keys())
    #     types = list(column_converted_type_dict.values())
    #     return cls(db1, name, columns, types, id_column_name=id_column_name)

    def initialize_columns(self):
        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND  TABLE_NAME  = '" + self.name + "';"
        columns = information_schema_table.select(query)
        self.columns = columns

    def convert_types_from_mysql(self):
        pass

    def initialize_types(self):
        self.types = self.get_all_types()

    def get_all_columns(self):
        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND TABLE_NAME  = '" + self.name + "'"
        columns = information_schema_table.select(query)

        return (columns)

    def get_all_types(self):

        data_types, data_lengths = self.get_data_types_and_character_lengths()
        for i in range(len(data_types)):
            if data_lengths[i] is not None:
                data_types[i] = data_types[i] + f"({data_lengths[i]})"
        return (data_types)


    """
        Returns a list of data types, where each element represents the category of the data ('varchar', 'int', etc.). 
        If a data type has an associated length, the length value will be included in a corresponding element of the
        data_lengths list, otherwise the element will have a None value. For example, 'nvarchar(2047)' would return
        'varchar' in the data_types list and 2047 in the data_lengths list.
    """
    def get_data_types_and_character_lengths(self):
        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
        query = f"SELECT DATA_TYPE,character_maximum_length FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND TABLE_NAME  = '" + self.name + "'"
        types = information_schema_table.select(query)
        data_types = [x[0] for x in types]
        data_lengths = [x[1] for x in types]

        return data_types,data_lengths

    def get_converted_python_types(self):
        SQL_TO_PYTHON = {v: k for k, v in PYTHON_TO_MYSQL_DATA_MAPPING.items()}
        python_types = []
        for type in self.types:
            if "varchar" in type:
                python_types.append("str")
            elif type in SQL_TO_PYTHON:
                python_types.append(SQL_TO_PYTHON[type])
            else:
                raise Exception("Unsupported type")

        return python_types

    def get_nullable_columns(self):
        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS where TABLE_SCHEMA = '{self.db1.DB_DATABASE}' and TABLE_NAME = '{self.name}' and IS_NULLABLE = 'YES'"
        nullable_columns = information_schema_table.select(query)
        return (nullable_columns)


    @classmethod
    def init_all_columns(cls, db1, name):
        temporary_table = cls(db1, name)
        columns = temporary_table.get_all_columns()
        types = temporary_table.get_all_types()

        if temporary_table.id_column_name in columns:
            id_col_index = columns.index(temporary_table.id_column_name)
            columns.pop(id_col_index)
            columns.insert(0, temporary_table.id_column_name)
            types.pop(id_col_index)
            types.insert(0, "int")



        return (cls(db1, name, columns, types))

    def get_last_id(self):
        """
        Returns the biggest id from table
        """

        last_id = self.select(f"SELECT {self.id_column_name} FROM {self.name} ORDER BY {self.id_column_name} DESC LIMIT 1;")

        return last_id[0][0]

    def get_num_of_records(self):
        """
        Returns the number of records in table
        """

        num_of_records = self.select(f"SELECT COUNT(*) FROM {self.name};")

        return num_of_records[0][0]

    def drop(self):
        query = "DROP TABLE " + self.name + ";"
        print(query)
        self.db1.execute(query)

    # @save_migration #TODO: Uncomment
    def create(self, foreign_keys=None):
        assert len(self.columns) == len(self.types)
        assert self.columns[0] == self.id_column_name
        assert self.types[0].lower() == "int"

        column_type_pairs = list(zip(self.columns, self.types))[1:]
        fields = ", ".join(
            [f"{column} {type_.upper()}" for column, type_ in column_type_pairs]
        )
        query = f"CREATE TABLE {self.name} ({self.id_column_name} INT UNSIGNED AUTO_INCREMENT PRIMARY KEY, {fields})"

        print(query)
        try:
            self.db1.execute(query)
        except Exception as e:
            print("Table " + self.name + " already exists:", e)
            print("Check the specification of table columns and their types")

    def insert(self, rows, batch=1, replace_apostrophes=True, try_mode=False, debug_mode=False, insert_id=False):
        start_index = 0 if insert_id else 1
        assert len(self.columns) == len(self.types)
        total_output=[]
        for k in range(len(rows)):
            if k % batch == 0:
                query = "INSERT INTO " + self.name + " ("
                for i in range(start_index, len(self.columns)):
                    if i < len(rows[k]) + 1:
                        # column name containing space needs to be wrapped in `...`, otherwise causes syntax error
                        if " " in self.columns[i]:
                            column_name = '`' + self.columns[i] + '`'
                        else:
                            column_name = self.columns[i]
                        query += column_name + ","
                if len(rows) < len(self.columns):
                    print(len(self.columns) - len(rows), "columns were not specified")
                if query[-1] == ',':
                    query = query[:-1]
                query = query + ") VALUES "

            query += "("
            for j in range(len(rows[k])):
                if rows[k][j] == "NULL" or rows[k][j] == "'NULL'" or rows[k][j] == None or rows[k][
                    j] == "None":  # NaN hodnoty
                    if "int" in self.types[j + start_index]:

                        if replace_apostrophes:
                            rows[k][j] = str(rows[k][j]).replace("'", "")
                        query += "NULL,"
                    else:
                        query += "NULL,"
                elif "nvarchar" in self.types[j + start_index]:
                    if replace_apostrophes:
                        rows[k][j] = str(rows[k][j]).replace("'", "")
                    query += "N'" + str(rows[k][j]) + "',"
                elif "varchar" in self.types[j + start_index]:
                    if replace_apostrophes:
                        rows[k][j] = str(rows[k][j]).replace("'", "")
                    query += "'" + str(rows[k][j]) + "',"
                elif "char" in self.types[j + start_index]:
                    if replace_apostrophes:
                        rows[k][j] = str(rows[k][j]).replace("'", "")
                    query += "'" + str(rows[k][j]) + "',"
                elif self.types[j + start_index] == "int":
                    query += str(rows[k][j]) + ","
                elif "datetime" in self.types[j + start_index]:
                    if replace_apostrophes:
                        rows[k][j] = str(rows[k][j]).replace("'", "")
                    query += "'" + str(rows[k][j]) + "',"
                elif "date" in self.types[j + start_index]:
                    query += "'" + str(rows[k][j]) + "',"
                elif "json" in self.types[j + start_index]:
                    query += f"'{rows[k][j]}', "


                else:
                    query += str(rows[k][j]) + ","
            if query[-1] == ",":
                query = query[:-1]

            query = query + "),"
            if k % batch == batch - 1 or k == len(rows) - 1:
                query = query[:-1]

                if debug_mode:
                    print(query)

                if not try_mode:
                    output=self.db1.execute(query)
                    total_output.append(output)
                else:
                    try:
                        output=self.db1.execute(query)
                        total_output.append(output)
                    except Exception as e:

                        print("Query", query, "Could not be inserted:", e)

                        # Write to logs only in debug mode
                        if debug_mode:
                            with open("log.txt", "a") as file:
                                file.write("Query " + str(query) + " could not be inserted:" + str(e) + "\n")
            
            elif len(total_output)==1:
                return(total_output[0])
            else:
                return(total_output)
                

    def add_foreign_key(self, foreign_key):
        parent_id = foreign_key['parent_id']
        parent = foreign_key['parent']
        query = "ALTER TABLE " + self.name + " MODIFY " + parent_id + " INT UNSIGNED"
        print(query)
        self.db1.execute(query)
        query = "ALTER TABLE " + self.name + " ADD FOREIGN KEY (" + parent_id + ") REFERENCES " + parent + "("+self.id_column_name+")"
        print(query)
        self.db1.execute(query)

    @save_migration
    def add_column(self, column_name, column_type):
        assert len(column_name) > 1
        command = "ALTER TABLE " + self.name + " ADD COLUMN " + column_name + " " + column_type
        try:
            self.db1.execute(command)
            self.columns.append(column_name)
            self.types.append(column_type)
        except Exception as e:
            print("Cant add column to table.")

    @save_migration
    def drop_column(self, column_name):
        assert len(column_name) > 1
        command = "ALTER TABLE " + self.name + " DROP COLUMN " + column_name
        try:
            print(command)
            self.db1.execute(command)
            index = self.columns.index(column_name)
            self.db1.columns.remove(column_name)
            self.db1.types.remove(self.db1.types[index])
        except Exception as e:
            print(e)
            print("Cant drop " + self.name)

    @save_migration
    def modify_column(self, column_name, new_column_type):
        assert len(column_name) > 1
        command = "ALTER TABLE " + self.name + " MODIFY COLUMN " + column_name + " " + new_column_type
        print(command)
        try:
            self.db1.execute(command)
            index = self.columns.index(column_name)
            self.types[index] = new_column_type
        except:
            print("Cant modify column to table.")


############### XLSX ##################

class XlsxTable(AbstractTable):
    def __init__(self, db1, name, columns=None, types=None, id_column_name = "id"):
        super().__init__(db1, name, columns, types)
        self.id_column_name = id_column_name

        table_filename = f"{self.name}.csv" if self.db1.is_csv else f"{self.name}.xlsx"
        self.table_directory_path: pathlib.Path = self.db1.db_directory_path / table_filename

    def _save_table(self, df: pd.DataFrame):
        if self.db1.is_csv:
            df.to_csv(self.table_directory_path, index=False)
        else:
            df.to_excel(self.table_directory_path, index=False)

    def create(self):
        if not self.table_directory_path.exists():
            df = pd.DataFrame(columns=self.columns)
            self._save_table(df)
        else:
            print(f"Table '{self.name}' already exists")

    def drop(self):
        self.table_directory_path.unlink(missing_ok=True)

    def select_to_df(self):
        # String must be typed with 'object' otherwise they get parsed arbitrarily and
        # then cast to string, e.g. 'true' -> True -> 'True'
        column_type_map = {
            column: object for column, type_ in self.column_type_dict.items() if type_ == "str"
        }
        date_columns = [
            column for column, type_ in self.column_type_dict.items() if type_ == "datetime"
        ]

        try:
            if self.db1.is_csv:
                df = pd.read_csv(
                    self.table_directory_path, dtype=column_type_map,
                    parse_dates=date_columns, encoding='utf-8'
                )
            else:
                df = pd.read_excel(
                    self.table_directory_path, dtype=column_type_map, 
                    parse_dates=date_columns
                )

            df.replace({np.nan: None}, inplace=True)
        except Exception as e:
            print(f"Error while reading data into XlsxTable: {e}")
            df = pd.DataFrame(columns=self.columns)

        return df

    def insert_from_df(self, df, batch=1, try_mode=False, debug_mode=False):
        assert len(df.columns) + 1 == len(self.columns)  # +1 because of id column

        original_df = self.select_to_df()

        # Need to save original dtypes, otherwise some of them will be changed during DF concatenation
        df_types = df.dtypes

        original_df.index = original_df[self.id_column_name]
        try:
            original_df = original_df.drop(original_df.columns[0], axis=1)
        except:
            print("First column could not be dropped")

        # handling nan values -> change to NULL TODO  
        for column in list(df.columns):
            df.loc[pd.isna(df[column]), column] = None

        def concat_with_reset_index_in_second_df(original_df, df):
            """Subsitute of reset_index method because we need to maintain the ids of original df"""
            if len(original_df.index) > 0:
                maximal_index = max(original_df.index)
            else:
                maximal_index = 0

            df.index = df.index + maximal_index + 1
            df.reindex(columns=original_df.columns.tolist())  # to ensure that columns get correctly inserted
            result_df = pd.concat([original_df, df]).astype(df_types)       # Cast to original dtypes

            return result_df

        df = concat_with_reset_index_in_second_df(original_df, df)
        df[self.id_column_name] = df.index
        df = df.reindex(columns=[self.id_column_name] + df.columns[:-1].tolist())  # uid as a first column
        df.reset_index(drop=True, inplace=True)

        self._save_table(df)

    def replace_from_df(self, df):
        assert len(df.columns) == len(self.columns)  # +1 because of id column
        #df.drop(df.columns[0], axis=1, inplace=True)
        self._save_table(df)

    # def update(self, variable_assign: str, where: Optional[str] = None):
    #     def split_assign(variable_assign):
    #         variable = variable_assign.split("=")[0]
    #         value = variable_assign.split("=")[1]
    #         try:
    #             value = int(value)  # integers
    #         except:
    #             value = value.split("'")[1]  # strings
    #         return (variable, value)

    #     variable, value = split_assign(variable_assign)
    #     df = self.select_to_df()
    #     print(where)
    #     print(variable, value)
    #     if where is None:
    #         df[variable] = value
    #         print(df)
    #     else:
    #         where_variable, where_value = split_assign(where)
    #         df[variable] = df[where_variable].apply(lambda x: value if str(x) == str(where_value) else x) #
    #     self.replace_from_df(df)

    def update(self, sql_column_update_string: str, sql_where_string: str) -> None:
        """Decompose provided parts of SQL UPDATE statement and update the xlsx file accordingly.

        TODO: Very fragile, unintuitive, and error-prone. Should be replaced with update_from_df().
        BUG: This will fail spectacularly on any values containing ',' in a
        `sql_column_update_string` or `sql_where_string'

        :param sql_column_update_string: e.g. "project_key='jakubatforloop.ai', project_name='Untitled Project'
        :type sql_column_update_string: str
        :param sql_where_string: e.g. "uid='1'"
        :type sql_where_string: str
        """

        # "project_key='jakubatforloop.ai', project_name='Untitled Project', last_active_pipeline_uid='1'"
        column_value_strings = sql_column_update_string.split(",")
        # ["project_key='jakubatforloop.ai'", " project_name='Untitled Project'", " last_active_pipeline_uid='1'"]
        column_value_pairs = [pair.split("=") for pair in column_value_strings]
        # [['project_key', "'jakubatforloop.ai'"], [' project_name', "'Untitled Project'"], [' last_active_pipeline_uid', "'1'"]]
        column_value_pairs = [
            [column.strip(' "\''), value.strip(' "\'')] for column, value in column_value_pairs
        ]
        # [['project_key', 'jakubatforloop.ai'], ['project_name', 'Untitled Project'], ['last_active_pipeline_uid', '1']]

        where_column, where_value = sql_where_string.split("=")
        where_value = where_value.strip(' "\'')

        df = self.select_to_df()
        columns_to_update = [pair[0] for pair in column_value_pairs]
        values_to_update = [pair[1] for pair in column_value_pairs]
        df.loc[df[where_column] == where_value, columns_to_update] = values_to_update

        self._save_table(df)

    def update_from_df(self, update_df: pd.DataFrame, where_column: str, where_value: Any) -> None:
        """Update the xlsx file with the provided dataframe.

        :param update_df: Dataframe with updated values - MUST only hold a single row
        :type update_df: pd.DataFrame
        :param where_column: Column name to use for WHERE clause.
        :type where_column: str
        :param where_value: Value to use for WHERE clause.
        :type where_value: Any
        """
        if not len(update_df) == 1:
            raise ValueError("There can only be one row in the UPDATE dataframe")

        table_df = self.select_to_df()
        table_df.loc[table_df[where_column] == where_value, update_df.columns] = update_df.values

        self._save_table(table_df)

    def delete(self, where=None) -> Optional[int]:
        def split_assign(variable_assign):
            variable = variable_assign.split("=")[0]
            value = variable_assign.split("=")[1]
            try:
                value = int(value)  # integers
            except:
                value = value.split("'")[1]  # strings
            return (variable, value)

        df = self.select_to_df()
        number_of_records = len(df)
        if where is None:
            df = df.iloc[0:0]
            deleted_count = number_of_records
        else:
            where_variable, where_value = split_assign(where)
            df.drop(df[df[where_variable] == where_value].index, inplace=True)
            deleted_count = number_of_records - len(df)
        self.replace_from_df(df)
        return deleted_count



import pandas as pd
from google.cloud import bigquery
from google.oauth2 import service_account


from dbhydra.src.abstract_db import AbstractDb
from dbhydra.src.tables import BigQueryTable


class BigQueryDb(AbstractDb):
    matching_table_class = BigQueryTable
    
    def __init__(self, db_details):
        self.credentials_path = db_details["DB_SERVER"]
        self.dataset = db_details["DB_DATABASE"]


        self.locally = False

        self.credentials = service_account.Credentials.from_service_account_file(self.credentials_path, scopes=[
            "https://www.googleapis.com/auth/cloud-platform"], )


    def connect_remotely(self):
        self.client = bigquery.Client(credentials=self.credentials, project=self.credentials.project_id)

    def connect_locally(self):
        raise Exception("Cannot connect locally to Big Query")


    def close_connection(self):
        self.client.close()

    def select_to_df(self, query):
        print(query)
        df = pd.read_gbq(query=query, project_id=self.project_id, credentials=self.credentials)

        return df

    def get_all_tables(self):
        query = f"""
        SELECT table_name
        FROM {self.dataset}.INFORMATION_SCHEMA.TABLES
        """
        rows = list(self.client.query(query))
        table_names = [row[0] for row in rows]
        return table_names

    def generate_table_dict(self):
        tables = self.get_all_tables()
        table_dict = dict()
        for i, table in enumerate(tables):
            table_dict[table] = BigQueryTable.init_all_columns(self, table)

        return (table_dict)




    def execute(self, query):
        query_job = self.client.query(query)
        results = query_job.result()  # Waits for job to complete.
        return results



import pymongo

from dbhydra.src.abstract_db import AbstractDb
from dbhydra.src.tables import MongoTable


class MongoDb(AbstractDb):
    matching_table_class = MongoTable

    def connect_locally(self):
        self.connection = pymongo.MongoClient(
            host=self.DB_SERVER,
            authSource=self.DB_DATABASE,
            username=self.DB_USERNAME,
            password=self.DB_PASSWORD,
            authMechanism='SCRAM-SHA-256'
        )

        self.database = self.connection.get_database(self.DB_DATABASE)
        print('Connected locally')

    def connect_remotely(self):
        self.connection = pymongo.MongoClient(
            "mongodb+srv://" + self.DB_USERNAME + ":" + self.DB_PASSWORD + "@" + self.DB_SERVER + "/" + self.DB_DATABASE + "?retryWrites=true&w=majority")
        print(self.connection.list_database_names())
        self.database = self.connection[self.DB_DATABASE]
        print(self.database)
        print("DB connection established")

    def execute(self, query, is_autocommitting=True):
        result=self.cursor.execute(query)
        if is_autocommitting:
            self.connection.commit()
        return(result)
    
    def get_all_tables(self):
        return self.database.list_collection_names()

    def create_table(self, name):
        table = self.database[name]
        return table

    def createTable(self, name):
        print("WARNING: `createTable` method will be deprecated in favor of `create_table`")
        return self.create_table(name)

    def close_connection(self):
        self.connection.close()
        print("DB connection closed.")

    def generate_table_dict(self):
        tables = self.get_all_tables()
        table_dict = dict()
        for i, table in enumerate(tables):
            table_dict[table] = MongoTable.init_all_columns(self, table)

        return (table_dict)



from dbhydra.src.abstract_db import AbstractDb
from dbhydra.src.tables import XlsxTable

import threading
import pathlib
import os


class XlsxDb(AbstractDb):
    """Folder-structure with .xlsx files representing database tables
    It does not need any server and runs locally with same syntax as other DB dialects
    """
    
    matching_table_class = XlsxTable
    
    def __init__(self, config_file="config.ini", db_details=None, is_csv=False):
        self.locally=True
        self.is_csv = is_csv
        if db_details is None:
            self.name="new_db"
            self.db_directory_path = None
        else:
            self.name = db_details.get("DB_DATABASE")
            self.db_directory_path = db_details.get("DB_DIRECTORY")

        self.lock = threading.Lock()
                
        if self.db_directory_path is None:
            self.db_directory_path = pathlib.Path(self.name) 
            
            
        self.python_database_type_mapping = {
        'int': "int",
        'float': "double",
        'str': "str",
        'tuple': "str",
        'list': "str",
        'dict': "str",
        'bool': "bool",
        'datetime': "datetime",
        'Jsonable': "str"
        }

        
        class DummyXlsxConnection: #compatibility with MySQL connection
            def begin(*args):
                pass
            def commit(*args):
                pass
            def rollback(*args):
                pass
            
        class DummyXlsxCursor: #compatibility with MySQL connection
            def execute(*args):    
                pass
            
            def fetchall(*args):
                pass
            
        self.cursor=DummyXlsxCursor()
        self.connection=DummyXlsxConnection()
        self.create_database()
        
    def connect_locally(self):
        pass #no real connection
        
    def connect_remotely(self):
        pass #no real connection

    def execute(self, query):
        pass
        # self.cursor.execute(query)
        # self.cursor.commit()

    def close_connection(self):
        pass
        # self.connection.close()
        # print("DB connection closed")

    def create_database(self):
        try:
            os.mkdir(self.db_directory_path)
            print("Database directory created")
        except FileExistsError:
            print("Database directory already exists")




class XlsxDB(XlsxDb):
    """Deprecated - do not remove until dbhydra 3.x"""
    def __init__(self, config_file="config.ini", db_details=None):
        print("Deprecation warning!, XlsxDB was renamed to XlsxDb and the old name will deprecated in future!")
        super().__init__(config_file=config_file, db_details=db_details)